/*
 *  Librera para el clculo de la FFT y la IFFT.
 *
 *  Avelino Herrera para "Slo Programadores"

	This software is free software; you can redistribute it and/or modify
	it under the terms of the GNU  General Public License as
	published by the Free Software Foundation; either version 2 of
	the License, or (at your option) any later version.
																																																																																																							 
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU  General Public License for more details.
																																																																																																							 
	You should have received a copy of the GNU  General Public
	License along with this software; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
	02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "fft.h"

/* Suma dos nmeros complejos */
void complex_add(complex *result, complex *x, complex *y) {
	result->re = x->re + y->re;
	result->im = x->im + y->im;
	return;
	}

/* Resta dos nmeros complejos */
void complex_sub(complex *result, complex *x, complex *y) {
	result->re = x->re - y->re;
	result->im = x->im - y->im;
	return;
	}

/* Multiplica dos nmeros complejos */
void complex_mul(complex *result, complex *x, complex *y) {
	result->re = (x->re * y->re) - (x->im * y->im);
	result->im = (x->re * y->im) + (x->im * y->re);
	return;
	}

/* Calcula la potencia natural de un nmero complejo */
void complex_pow(complex *result, complex *base, int exponent) {
	complex temp;
	int i;

	result->re = 1.0;
	result->im = 0.0;
	for (i = 0; i < exponent; i++) {
		temp = *result;
		complex_mul(result, base, &temp);
		}
	return;
	}

/* Calcula el valor absoluto de un complejo */
float complex_abs(complex *v) {
	return (float) sqrt((v->re * v->re) + (v->im * v->im));
	}

/* Calcula la representacin inversa de bits del nmero "a" (n bits) */
int bit_wise(int a, int n) {
	int v, r, i;

	v = 1 << (n - 1);
	r = 0;
	for (i = 0; i < n; i++) {
		if (((a >> i) & 1) == 1)
			r |= v >> i;
		}
	return r;
	}

/* Calcula la FFT sobre un array de 2^ln elementos */
void fft(complex *vector, int ln) {
	int n, level, i, j, jump, start;
	complex temp, temp1, mid, base, w;

	n = 1 << ln;
	mid.re = 0.5;
	mid.im = 0.0;
	/* reordenamos el array mediante la funcin bit_wise */
	for (i = 0; i < (n >> 1); i++) {
		j = bit_wise(i, ln);
		temp = vector[i];
		vector[i] = vector[j];
		vector[j] = temp;
		}
	/* calculamos la fft */
	for (level = 0; level < ln; level++) {
		jump = 1 << level;
		base.re = cos(M_PI / jump);
		base.im = -sin(M_PI / jump);
		for (start = 0; start < jump; start++) {
			complex_pow(&w, &base, start);
			for (i = start; i < n; i += (jump << 1)) {
				j = i + jump;
				complex_mul(&temp, &vector[j], &w);

				complex_sub(&temp1, &vector[i], &temp);
				complex_mul(&vector[j], &temp1, &mid);

				complex_add(&temp1, &vector[i], &temp);
				complex_mul(&vector[i], &temp1, &mid);
				}
			}
		}
	return;
	}

/* Calcula la IFFT sobre un array de 2^ln elementos */
void ifft(complex *vector, int ln) {
	int n, level, i, j, jump, start;
	complex temp, temp1, mid, base, w;

	n = 1 << ln;
	mid.re = 0.5;
	mid.im = 0.0;
	/* reodenamos el array mediante la funcin bit_wise */
	for (i = 0; i < (n >> 1); i++) {
		j = bit_wise(i, ln);
		temp = vector[i];
		vector[i] = vector[j];
		vector[j] = temp;
		}
	/* calculamos la IFFT */
	for (level = 0; level < ln; level++) {
		jump = 1 << level;
		base.re = cos(M_PI / jump);
		/* igual que la FFT pero con base.im positivo */
		base.im = sin(M_PI / jump);
		for (start = 0; start < jump; start++) {
			complex_pow(&w, &base, start);
			for (i = start; i < n; i += (jump << 1)) {
				j = i + jump;
				complex_mul(&temp, &vector[j], &w);

				complex_sub(&temp1, &vector[i], &temp);
				complex_mul(&vector[j], &temp1, &mid);

				complex_add(&temp1, &vector[i], &temp);
				complex_mul(&vector[i], &temp1, &mid);
				}
			}
		}
	/* hay que escalar los valores resultantes del array */
	for (i = 0; i < n; i++) {
		vector[i].re *= n;
		vector[i].im *= n;
		}
	return;
	}

/* Devuelve el logaritmo en base 2 del parmetro */
int log2i(int n) {
	return (int) rint(log(n) / log(2.0));
	}

