/*
 * (c) Avelino Herrera Morales
 *
 * Mini-Sampler basado en el estándar de instrumentos XI (eXtended Instrument, de FastTracker).
 * Este programa lee de memoria compartida los eventos MIDI previamente parseados.
 * El programa "midiread.c" se debe ejecutar en otra consola antes que "readxi.c", ya que el primero crea la zona de
 * memoria compartida a la cual se conecta el segundo. Los eventos MIDI perseados por "midiread.c" son traducidos a
 * notas musicales por este programa.
 *
	This software is free software; you can redistribute it and/or modify
	it under the terms of the GNU  General Public License as
	published by the Free Software Foundation; either version 2 of
	the License, or (at your option) any later version.
																																																																																																							 
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU  General Public License for more details.
																																																																																																							 
	You should have received a copy of the GNU  General Public
	License along with this software; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
	02111-1307, USA.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/soundcard.h>

#include <sys/ipc.h>
#include <sys/shm.h>
#include "shm_midi.h"   /* este programa lee datos MIDI de la memoria compartida siguiendo la estructura de shm_midi.h */


#define  MAX_ENVELOPE_POINTS  12

#define  ENVELOPE_ON       1
#define  ENVELOPE_SUSTAIN  2
#define  ENVELOPE_LOOP     4

#define  SAMPLE_16_BIT         0x10
#define  SAMPLE_LOOP_FORWARD   1
#define  SAMPLE_LOOP_PINGPONG  2

#define  SAMPLE_RATE     44100
#define  TICK_SAMPLES    (SAMPLE_RATE / 36)


typedef struct {
	char name[23];

	unsigned char volume;
	signed char fine_tune;       /*  Octavas de semitono  */
	unsigned char panning;
	signed char rel_note;

	unsigned int length;
	unsigned int loop_start;
	unsigned int loop_length;
	unsigned char flags;

	signed char *data8;
	signed short int *data16;
	} ft_sample;

typedef struct {
	unsigned short int pos;
	unsigned short int val;
	} ft_envelope_point;

typedef struct {
	ft_envelope_point point[MAX_ENVELOPE_POINTS];
	unsigned char num_points;
	unsigned char sustain_point;
	unsigned char loop_start_point;
	unsigned char loop_end_point;
	unsigned char flags;
	} ft_envelope;

typedef struct {
	char name[24];

	ft_envelope vol_env;
	ft_envelope pan_env;

	unsigned char vibrato_type;
	unsigned short int vibrato_rate;
	unsigned short int vibrato_depth;
	unsigned short int vibrato_sweep;

	unsigned int volume_fade;

	unsigned char sample_map[96];
	unsigned short int num_samples;
	ft_sample *sample;
	} ft_instrument;

typedef struct {
  	ft_envelope *env;
	int prev_point;
	int next_point;
	unsigned long t;
  	} ft_running_envelope;

typedef struct {
  	ft_instrument *ins;
  	ft_running_envelope vol_renv;
	ft_running_envelope pan_renv;
	ft_sample *sample;
	double t;
	double t_inc;
	float last_volume_value;
	float fadeout_volume;
	int to_fade;
	int forward;
	int press;
  	} ft_running_instrument;


/****  Playing  ****/
#define  NUM_CHANNELS   32
#define  BUFFER_SIZE    512
#define  AMPLIFICATION  0.5

shm_midi *midi_mem;

typedef struct {
	ft_running_instrument rins;
	int active;
	int key;
	float amplification;
	} channel;

channel ch[NUM_CHANNELS];
int key_status[128];

#undef    __USE_FILTER__
#define    __USE_INTERPOLATION__

#ifdef  __USE_FILTER__
	float freq = 0.3, reso = 0.2;
	float low = 0.0, mid = 0.0, high = 0.0;
#endif



void ft_running_envelope_start(ft_running_envelope *r) {
  	r->prev_point = -1;
	if (r->env->num_points > 0)
		r->next_point = 0;
	 else
	   	r->next_point = -1;
	r->t = 0;
	return;
  	}

void ft_running_envelope_work(ft_running_envelope *r, int press,
                              float *value, int size, int *last_buffer) {
  	int i;
	float ppos, npos, pval, nval, rel_pos;

	*last_buffer = 0;
	if (!(r->env->flags & ENVELOPE_ON)) {
		if (press) {
			for (i = 0; i < size; i++)
				value[i] = -256;
			}
		 else {
			for (i = 0; i < size; i++)
				value[i] = 0;
			*last_buffer = 1;
			}
		return;
		}
	for (i = 0; i < size; i++) {
		ppos = (r->prev_point < 0) ? 0 : r->env->point[r->prev_point].pos;
		pval = (r->prev_point < 0) ? 0 : r->env->point[r->prev_point].val;
		npos = r->env->point[r->next_point].pos;
		nval = r->env->point[r->next_point].val;
		ppos *= TICK_SAMPLES;
		npos *= TICK_SAMPLES;
		if (npos == ppos)
			value[i] = nval;
		 else {
			rel_pos = (r->t - ppos) / (npos - ppos);
			rel_pos = (rel_pos > 1) ? 1 : rel_pos;
			value[i] = pval + (rel_pos * (nval - pval));
			}
		if (r->t == npos) {
			if ((r->env->flags & ENVELOPE_LOOP) &&
			    (r->next_point == r->env->loop_end_point) && press) {
				if (r->env->loop_start_point == r->env->loop_end_point)
					r->next_point = r->env->loop_start_point;
				 else
					r->next_point = r->env->loop_start_point + 1;
				r->prev_point = r->env->loop_start_point;
				r->t = r->env->point[r->prev_point].pos * TICK_SAMPLES;
				}
			else if ((r->env->flags & ENVELOPE_SUSTAIN) &&
			         (r->next_point == r->env->sustain_point) && press) {
				r->prev_point = r->next_point = r->env->sustain_point;
				r->t = r->env->point[r->prev_point].pos * TICK_SAMPLES;
				}
			else if (r->next_point == (r->env->num_points - 1)) {
				r->prev_point = r->next_point;
				r->t = r->env->point[r->prev_point].pos * TICK_SAMPLES;
				if (!press)
					*last_buffer = 1;
				}
			else {
				if (r->prev_point < r->next_point)
					r->prev_point++;
				r->next_point++;
				r->t = r->env->point[r->prev_point].pos * TICK_SAMPLES;
				}
			}
		 else
			r->t++;
	  	}
	return;
	}

void ft_running_envelope_test(void) {
  	ft_running_envelope renv;
	ft_envelope env;
	float val[10];
	int i, j, last;

	env.num_points = 4;
	env.point[0].pos = 0;
	env.point[0].val = 0;
	env.point[1].pos = 4;
	env.point[1].val = 10;
	env.point[2].pos = 7;
	env.point[2].val = 2;
	env.point[3].pos = 10;
	env.point[3].val = 1;
	env.loop_start_point = 1;
	env.loop_end_point = 3;
	env.sustain_point = 1;
	env.flags = ENVELOPE_ON | ENVELOPE_SUSTAIN | ENVELOPE_LOOP;
	renv.env = &env;
	ft_running_envelope_start(&renv);
	for (i = 0; i < 1000; i++) {
	  	if (i < 20)
		  	ft_running_envelope_work(&renv, 1, val, 10, &last);
		 else
		  	ft_running_envelope_work(&renv, 0, val, 10, &last);
		for (j = 0; j < 10; j++)
		  	printf("%f\n", val[j]);
		}
	return;
  	}

void ft_running_instrument_start(ft_running_instrument *rins,
	                             ft_instrument *ins, int note, int press) {
  	float frequency, period;

	rins->ins = ins;
	/*  Init volume and panning envelopes  */
	rins->vol_renv.env = &(ins->vol_env);
	rins->pan_renv.env = &(ins->pan_env);
	ft_running_envelope_start(&(rins->vol_renv));
	ft_running_envelope_start(&(rins->pan_renv));
	/*  Test keyboard limits  */
	if (note < 0)
	  	note = 0;
	else if (note > 95)
	  	note = 95;
	/*  Determine sample  */
	rins->sample = &(ins->sample[ins->sample_map[note]]);
	/*  Determine playback frequency  */
	note += rins->sample->rel_note;
	period = (10.0 * 12.0 * 16.0 * 4.0) - (note * 16.0 * 4.0) -
	         (rins->sample->fine_tune / 2.0);
	frequency = 8363 * pow(2.0, ((6 * 12 * 16 * 4) - period) / (12 * 16 * 4));
	rins->t = 0;
	rins->t_inc = frequency / SAMPLE_RATE;
	rins->forward = 1;
	rins->press = press;
	rins->fadeout_volume = (float) ins->volume_fade;
	rins->to_fade = 0;
	rins->last_volume_value = 0;
	return;
  	}

void ft_running_instrument_work(ft_running_instrument *rins,
	                            float *output, int *last) {
  	float vol[BUFFER_SIZE], pan[BUFFER_SIZE];
	float smp, fade;
#ifdef  __USE_INTERPOLATION__
	float smp1, smp2;
#endif
	int i;
	ft_sample *s;
	int vlast, plast;

	s = rins->sample;
	rins->to_fade = !rins->press;
	if (!(rins->to_fade)) {
		ft_running_envelope_work(&(rins->vol_renv), rins->press, vol, BUFFER_SIZE, &vlast);
		ft_running_envelope_work(&(rins->pan_renv), rins->press, pan, BUFFER_SIZE, &plast);
		}
	 else {
		for (i = 0; i < BUFFER_SIZE; i++) {
			/*
			vol[i] = rins->last_volume_value - (rins->fadeout_volume / 16384.0);
			*/
			fade = (float) i / BUFFER_SIZE;
			vol[i] = rins->last_volume_value - (fade * rins->fadeout_volume /
			         16384.0);
			vol[i] *= 64.0;
			if (vol[i] < 0) {
				vol[i] = 0;
				*last = 1;
				}
			}
		rins->last_volume_value = vol[BUFFER_SIZE - 1] / 64.0;
		}
	for (i = 0; i < BUFFER_SIZE; i++) {
	  	/*  Scale volume to 0..1  */
		if (vol[i] < 0)
			vol[i] = 1;
		 else
			vol[i] /= 64.0;
		/*  Scale panning to -1..1  */
		if (pan[i] < -1)
			pan[i] = 0;
		 else
			pan[i] = ((pan[i] / 64.0) * 2) - 1;
		/*  Loop control  */
		if ( ((s->flags & 0x0F) == SAMPLE_LOOP_FORWARD) &&
			      (rins->t >= s->loop_start + s->loop_length) ) {
		  	rins->t -= s->loop_length;
			}
		if ( ((s->flags & 0x0F) == SAMPLE_LOOP_PINGPONG) &&
			 (rins->t >= s->loop_start + s->loop_length) &&
			  rins->forward ) {
		  	rins->forward = 0;
			}
		if ( ((s->flags & 0x0F) == SAMPLE_LOOP_PINGPONG) &&
			 (rins->t <= s->loop_start) &&
			 (!rins->forward) ) {
			rins->forward = 1;
			}
		/*  Generate output  */
#ifdef  __USE_INTERPOLATION__
		if (rins->t >= s->length)
		  	smp = 0;
		else if (rins->sample->flags & SAMPLE_16_BIT) {
		  	smp1 = rins->sample->data16[(int) floor(rins->t)] / 32768.0;
			if ((int) ceil(rins->t) >= s->length)
				smp2 = smp1;
			 else
			  	smp2 = rins->sample->data16[(int) ceil(rins->t)] / 32768.0;
			}
		else {
		   	smp1 = rins->sample->data8[(int) floor(rins->t)] / 128.0;
			if ((int) ceil(rins->t) >= s->length)
				smp2 = smp1;
			 else
			  	smp2 = rins->sample->data8[(int) ceil(rins->t)] / 128.0;
			}
		smp = smp1 + (rins->t - (int) floor(rins->t)) * (smp2 - smp1);
#else
		if (rins->t >= s->length)
		  	smp = 0;
		else if (rins->sample->flags & SAMPLE_16_BIT)
		  	smp = rins->sample->data16[(int) rint(rins->t)] / 32768.0;
		else
		   	smp = rins->sample->data8[(int) rint(rins->t)] / 128.0;
#endif
		output[i] = vol[i] * smp;
		/*  Increment sample index by note  */
		if (rins->forward)
			rins->t += rins->t_inc;
		 else
		   	rins->t -= rins->t_inc;
	  	}
	rins->last_volume_value = vol[BUFFER_SIZE - 1];
	if (vlast && (rins->to_fade == 0)) {
		rins->last_volume_value = vol[BUFFER_SIZE - 1];
		rins->to_fade = 1;
		}
	return;
  	}

int read_ft_instrument(char *filename, ft_instrument *fti) {
	FILE *f;
	unsigned char a[29], b[38];
	int i, j, trash, p;
	signed short int *d16;
	signed char *d8;

	/*  Open file  */
	f = fopen(filename, "rb");
	if (f == NULL)
		return 0;

	/*  Read signature  */
	fread(a, 1, 21, f);
	a[21] = 0;
	if (strcmp(a, "Extended Instrument: ") != 0)
		return 0;

	/*  Instrument name  */
	fread(a, 1, 22, f);
	strncpy(fti->name, a, 22);

	/*  Instrument version  */
	fread(a, 1, 23, f);

	/*  Sample map  */
	fread(fti->sample_map, 1, 96, f);

	/*  Volume / Panning envelope points  */
	fread(fti->vol_env.point, 1, 48, f);
	fread(fti->pan_env.point, 1, 48, f);

	/*  Envelope parameters  */
	fread(b, 1, 16, f);
	fti->vol_env.num_points = b[0];
	fti->vol_env.sustain_point = b[2];
	fti->vol_env.loop_start_point = b[3];
	fti->vol_env.loop_end_point = b[4];
	fti->vol_env.flags = b[8];

	fti->pan_env.num_points = b[1];
	fti->pan_env.sustain_point = b[5];
	fti->pan_env.loop_start_point = b[6];
	fti->pan_env.loop_end_point = b[7];
	fti->pan_env.flags = b[9];

	fti->vibrato_type = b[10];
	fti->vibrato_rate = b[13];
	fti->vibrato_depth = b[12];
	fti->vibrato_sweep = b[11];

	fti->volume_fade = ((unsigned short int)b[15] << 8) | b[14];

	fread(a, 1, 24, f);
	fti->num_samples = ((unsigned short int)a[23] << 8) | a[22];

	/*  Load samples  */
	fti->sample = (ft_sample *) malloc(sizeof(ft_sample) * fti->num_samples);
	if (fti->sample == NULL)
		return 0;
	for (i = 0; i < fti->num_samples; i++) {
	  	fread(&(fti->sample[i].length), 1, 4, f);
		fread(&(fti->sample[i].loop_start), 1, 4, f);
		fread(&(fti->sample[i].loop_length), 1, 4, f);
		fread(&(fti->sample[i].volume), 1, 1, f);
		fread(&(fti->sample[i].fine_tune), 1, 1, f);
		fread(&(fti->sample[i].flags), 1, 1, f);
		fread(&(fti->sample[i].panning), 1, 1, f);
		fread(&(fti->sample[i].rel_note), 1, 1, f);
		fread(&trash, 1, 1, f);
		fread(fti->sample[i].name, 1, 22, f);
		fti->sample[i].name[22] = 0;
		}

	/*  Load sample data  */
	for (i = 0; i < fti->num_samples; i++) {
		if (fti->sample[i].length == 0) {
			fti->sample[i].data8 = (signed char *) fti->sample[i].data16 = NULL;
			continue;
			}
		if (fti->sample[i].flags & SAMPLE_16_BIT) {
			fti->sample[i].length >>= 1;
			fti->sample[i].loop_start >>= 1;
			fti->sample[i].loop_length >>= 1;
			d16 = fti->sample[i].data16 = malloc(2 * fti->sample[i].length);
			if (d16 == NULL)
				return 0;
			fread(d16, 1, 2 * fti->sample[i].length, f);
			for (j = fti->sample[i].length, p = 0; j; j--) {
				p += *d16;
				*d16++ = p;
				}
			}
		 else {
			d16 = (signed short int *) fti->sample[i].data8 = malloc(2 * fti->sample[i].length);
			if (d16 == NULL)
				return 0;
			d8 = (signed char *) d16;
			d8 += fti->sample[i].length;
			fread(d8, 1, fti->sample[i].length, f);
			for (j = fti->sample[i].length, p = 0; j; j--) {
				p += *d8++;
				*d16++ = p << 8;
				}
			}
		}

	return 1;
	}


void show_ft_instrument(ft_instrument *fti) {
	int i;

	printf("*** Instrument ***\n");
	printf("Name : %s\n", fti->name);
	printf("Sample map :\n");
	for (i = 0; i < 96; i++)
		printf("%2u  ", (unsigned int) fti->sample_map[i]);
	printf("\n");
	printf("Volume envelope (%d points) :\n", fti->vol_env.num_points);
	for (i = 0; i < fti->vol_env.num_points; i++)
		printf("pos=%3d val=%3d\n",
		       fti->vol_env.point[i].pos, fti->vol_env.point[i].val);
	printf("\tSustain point : %d\n", fti->vol_env.sustain_point);
	printf("\tLoop start point : %d\n", fti->vol_env.loop_start_point);
	printf("\tLoop end point : %d\n", fti->vol_env.loop_end_point);
	printf("\tFlags : ");
	if (fti->vol_env.flags & ENVELOPE_ON)
		printf("ON  ");
	if (fti->vol_env.flags & ENVELOPE_SUSTAIN)
		printf("SUSTAIN  ");
	if (fti->vol_env.flags & ENVELOPE_LOOP)
		printf("LOOP");
	printf("\n");
	printf("Panning envelope (%d points):\n", fti->pan_env.num_points);
	for (i = 0; i < fti->pan_env.num_points; i++)
		printf("pos=%3d val=%3d\n",
		       fti->pan_env.point[i].pos, fti->pan_env.point[i].val);
	printf("\tSustain point : %d\n", fti->pan_env.sustain_point);
	printf("\tLoop start point : %d\n", fti->pan_env.loop_start_point);
	printf("\tLoop end point : %d\n", fti->pan_env.loop_end_point);
	printf("\tFlags : ");
	if (fti->pan_env.flags & ENVELOPE_ON)
		printf("ON  ");
	if (fti->pan_env.flags & ENVELOPE_SUSTAIN)
		printf("SUSTAIN  ");
	if (fti->pan_env.flags & ENVELOPE_LOOP)
		printf("LOOP");
	printf("\n");
	printf("Vibrato :\n");
	printf("\tType : ");
	if (fti->vibrato_type == 0)
		printf("SINE\n");
	else if (fti->vibrato_type == 1)
		printf("SQUARE\n");
	else if (fti->vibrato_type == 2)
		printf("SAWTOOTH UP\n");
	else if (fti->vibrato_type == 3)
		printf("SAWTOOTH DOWN\n");
	else
		printf("INVALID\n");
	printf("\tRate : %d\n", fti->vibrato_rate);
	printf("\tDepth : %d\n", fti->vibrato_depth);
	printf("\tSweep : %d\n", fti->vibrato_sweep);
	printf("Volume fade : %d\n", fti->volume_fade);
	printf("Number of samples : %d\n", fti->num_samples);
	printf("*** SAMPLES ***\n");
	for (i = 0; i < fti->num_samples; i++) {
		printf("Sample %d\n", i);
		printf("\tName : %s\n", fti->sample[i].name);
		printf("\tVolume : %d\n", fti->sample[i].volume);
		printf("\tPanning : %d\n", fti->sample[i].panning);
		printf("\tFine tune : %d\n", fti->sample[i].fine_tune);
		printf("\tRelative Note : %d\n", fti->sample[i].rel_note);
		printf("\tFlags : ");
		if (fti->sample[i].flags & SAMPLE_16_BIT)
			printf("16 BIT   ");
		if ((fti->sample[i].flags & 0x0F) == SAMPLE_LOOP_FORWARD)
			printf("FORWARD LOOP");
		else if ((fti->sample[i].flags & 0x0F) == SAMPLE_LOOP_PINGPONG)
			printf("PINGPONG LOOP");
		printf("\n");
		printf("\tLength : %d\n", fti->sample[i].length);
		printf("\tLoop start : %d\n", fti->sample[i].loop_start);
		printf("\tLoop length : %d\n", fti->sample[i].loop_length);
		}
	return;
	}

int log2i(int x) {
	return (int) rint(log((double) x) / log(2.0));
	}

int audio_open(void) {
  	int stereo = 1;
	int format = AFMT_S16_LE;
	int sample_rate = SAMPLE_RATE;
	int f;
	int setting = 0x00040000;

	if ((f = open("/dev/dsp", O_WRONLY, 0)) == -1) {
	  	perror("audio_open");
		exit(-1);
	  	}
	setting |= log2i(BUFFER_SIZE * sizeof(signed short int) * 2);
	printf("Fragment setting = %8X\n", setting);
	ioctl(f, SNDCTL_DSP_SETFRAGMENT, &setting);
	ioctl(f, SNDCTL_DSP_SETFMT, &format);
	ioctl(f, SNDCTL_DSP_STEREO, &stereo);
	ioctl(f, SNDCTL_DSP_SPEED, &sample_rate);
	return f;
  	}

void audio_play_8(int fd, signed char *buffer) {
  	signed short int p[BUFFER_SIZE];
	int i;

	for (i = 0; i < BUFFER_SIZE; i++)
	  	p[i] = buffer[i];
	write(fd, p, sizeof(signed short int) * BUFFER_SIZE);
	return;
  	}

void audio_play_16(int fd, signed short int *buffer) {
  	write(fd, buffer, sizeof(signed short int) * BUFFER_SIZE);
	return;
  	}

void audio_play_f(int fd, float *left_output, float *right_output) {
  	signed short int p[BUFFER_SIZE * 2];
	signed short int background[BUFFER_SIZE * 2];
	int i;
	register int temp;

	read(0, background, BUFFER_SIZE * sizeof(signed short int) * 2);
	for (i = 0; i < BUFFER_SIZE; i++) {
		temp = (int) rint(left_output[i] * 32767 + background[i * 2]);
		if (temp > 32767)
			p[i * 2] = 32767;
		else if (temp < -32768)
			p[i * 2] = -32768;
		else
			p[i * 2] = (signed short int) temp;
		temp = (int) rint(right_output[i] * 32767 + background[(i * 2) + 1]);
		if (temp > 32767)
			p[(i * 2) + 1] = 32767;
		else if (temp < -32768)
			p[(i * 2) + 1] = -32768;
		else
			p[(i * 2) + 1] = (signed short int) temp;
		}
	/* ioctl(fd, SNDCTL_DSP_SYNC, 0); */
	write(fd, p, BUFFER_SIZE * 2 * sizeof(signed short int));
	return;
  	}


int get_free_channel(channel *c) {
	int i;

	for (i = 0; i < NUM_CHANNELS; i++) {
		if (!c[i].active)
			return i;
		}
	return -1;
	}


int main(int argc, char *argv[]) {
	ft_instrument i1;
	ft_running_instrument ri1;
	int dsp, i, j, last_buffer = 0;
	float aux_output[BUFFER_SIZE];
	float output[BUFFER_SIZE];
	int shmid;
	audio_buf_info dsp_info;
	int block_size;


	read_ft_instrument(argv[1], &i1);
	show_ft_instrument(&i1);

	dsp = audio_open();
	ioctl(dsp, SNDCTL_DSP_GETBLKSIZE, &block_size);
	printf("oss block size = %d\n", block_size);
	ioctl(dsp, SNDCTL_DSP_GETOSPACE, &dsp_info);
	printf("oss output fragments = %d\n", dsp_info.fragments);
	printf("oss output allocated fragments = %d\n", dsp_info.fragstotal);
	printf("oss output fragment size = %d\n", dsp_info.fragsize);
	printf("oss output available bytes = %d\n", dsp_info.bytes);

	
	/*  Abrimos la memoria compartida para recoger los datos midi  */
	shmid = shmget(SHM_MIDI_KEY, sizeof(shm_midi) * SHM_MIDI_CH,
	               0700 | IPC_CREAT);
	if (shmid < 0) {
		perror("main");
		return -1;
		}
	midi_mem = shmat(shmid, (const void *) 0, SHM_RND);
	/***************************************************************/

	for (i = 0; i < NUM_CHANNELS; i++)
		ch[i].active = 0;
	for (i = 0; i < 128; i++)
		key_status[i] = -1;
	while (1) {
		/*  midi read  */
		for (i = 0; i < 128; i++) {
			if (midi_mem[0].keyboard_hit[i] && (key_status[i] < 0)) {
				key_status[i] = get_free_channel(ch);
				if (key_status[i] < 0)
					break;
				ft_running_instrument_start(&ch[key_status[i]].rins, &i1,
				                            i, 1);
				ch[key_status[i]].active = 1;
				ch[key_status[i]].key = i;
				ch[key_status[i]].amplification = (float) midi_mem[0].keyboard_velocity[i] / 127.0;
				}
			else if (!midi_mem[0].keyboard_hit[i] && (key_status[i] >= 0)) {
				ch[key_status[i]].rins.press = 0;
				key_status[i] = -1;
				}
			}

		/*  sound output */
		memset(output, 0, BUFFER_SIZE * sizeof(float));
		for (i = 0; i < NUM_CHANNELS; i++) {
			if (!ch[i].active)
				continue;
			ft_running_instrument_work(&ch[i].rins, aux_output, &last_buffer);
			if (last_buffer) {
				ch[i].active = 0;
				last_buffer = 0;
				}
			for (j = 0; j < BUFFER_SIZE; j++)
				output[j] += AMPLIFICATION * aux_output[j] * ch[i].amplification * ((float) midi_mem[0].controller[7] / 127.0);
			}

#ifdef  __USE_FILTER__
		freq = (float) midi_mem[0].controller[1] / 127.0;
		for (i = 0; i < BUFFER_SIZE; i++) {
			high = output[i] - (reso * mid) - low;
			mid += high * freq;
			low += mid * freq;
			output[i] = low;
			}
#endif
		audio_play_f(dsp, output, output);
		}

	close(dsp);
	return 0;
	}
