/*
	Un generador de cuerda frotada.

	Avelino Herrera Morales para "Slo Programadores".

	This software is free software; you can redistribute it and/or modify
	it under the terms of the GNU  General Public License as
	published by the Free Software Foundation; either version 2 of
	the License, or (at your option) any later version.
																																																																																																							 
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU  General Public License for more details.
																																																																																																							 
	You should have received a copy of the GNU  General Public
	License along with this software; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
	02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>


#define  BUFFER_SIZE           256
#define  SAMPLING_RATE         44100
#define  LOWEST_MUSIC_FREQ     20.0
/* Determinamos el tamao mximo de la gua a partir de la frecuencia audible
   ms baja */
#define  MAX_WAVEGUIDE_SIZE    ((int)((1.0/LOWEST_MUSIC_FREQ) * SAMPLING_RATE/2.0))
#define  BOW_TABLE_SIZE        1024
#define  FILTER_CONSTANT       -0.642



typedef struct {
	float freq;
	int waveguide_size;
	float forward[MAX_WAVEGUIDE_SIZE];
	float backward[MAX_WAVEGUIDE_SIZE];
	int waveguide_size1, waveguide_size2;
	float bow_table[BOW_TABLE_SIZE];
	int split;
	float bow_velocity;
	int start;
	float filter_out;
	float output;
	} bowed_string;



int get_bow_table_index(float x) {
	x += 1.0;
	x = (x > 2.0) ? 2.0 : ((x < 0.0) ? 0.0 : x);
	return (int) rint(x * (BOW_TABLE_SIZE - 1) / 2.0);
	}


/******************************************************************************
	Inicializamos los campos de la estructura "bowed_string" para tocar.
*/
void bowed_string_init(bowed_string *p, float freq, float bow_point,
                       float bow_velocity) {
	int i;

	p->waveguide_size = (int) rint((1.0 / freq) * SAMPLING_RATE / 2.0);
	p->split = (int) rint(bow_point * (p->waveguide_size - 1));
	p->bow_velocity = bow_velocity;
	p->waveguide_size1 = p->split;
	p->waveguide_size2 = p->waveguide_size - p->waveguide_size1;
	for (i = 0; i < BOW_TABLE_SIZE; i++) {
		p->bow_table[i] = (float)i * 2.0 / (BOW_TABLE_SIZE - 1);
		if ((p->bow_table[i] < 0.3) || (p->bow_table[i] >= 1.6))
			p->bow_table[i] = 0.0;
		else if ((p->bow_table[i] >= 0.6) && (p->bow_table[i] < 1.3))
			p->bow_table[i] = 1.0;
		else if ((p->bow_table[i] >= 0.3) && (p->bow_table[i] < 0.6))
			p->bow_table[i] = (p->bow_table[i] - 0.3) / 0.3;
		else if ((p->bow_table[i] >= 1.3) && (p->bow_table[i] < 1.6))
			p->bow_table[i] = 1.0 - ((p->bow_table[i] - 1.3) / 0.3);
		}
	p->filter_out = 0;
	p->start = 1;
	return;
	}


/******************************************************************************
	Calcula BUFFER_SIZE muestras de la cuerda frotada.
*/
void bowed_string_work(bowed_string *p, float *output) {
	int i, k;
	float forward_out, backward_out;
	float forward_bow_out, backward_bow_out;
	float add, mul;

	for (k = 0; k < BUFFER_SIZE; k++) {
		if (p->start) {
			/* Condicin inicial de la cuerda */
			for (i = 0; i < p->split; i++)
				p->forward[i] = (float)i / p->split;
			for (i = p->split; i < p->waveguide_size; i++)
				p->forward[i] = 1.0 - ((float)(i - p->split) / p->waveguide_size2);
			for (i = 0; i < p->waveguide_size; i++)
				p->backward[i] = -p->forward[i];
			p->filter_out = 0;
			p->start = 0;
			}
		else {
			/* Desplazamos las guas */
			forward_out = p->forward[p->waveguide_size - 1];
			memmove(&(p->forward[p->split + 1]), &(p->forward[p->split]),
			        (p->waveguide_size2 - 1) * sizeof(float));
			backward_bow_out = p->backward[p->split];
			memmove(&(p->backward[p->split]), &(p->backward[p->split + 1]),
			        (p->waveguide_size2 - 1) * sizeof(float));
			forward_bow_out = p->forward[p->split - 1];
			memmove(&(p->forward[1]), &(p->forward[0]),
			        (p->waveguide_size1 - 1) * sizeof(float));
			backward_out = p->backward[0];
			memmove(&(p->backward[0]), &(p->backward[1]),
			        (p->waveguide_size1 - 1) * sizeof(float));
			/* Un extremo de la cuerda */
			p->filter_out = (forward_out * (1 + FILTER_CONSTANT)) -
			                (FILTER_CONSTANT * p->filter_out);
			p->backward[p->waveguide_size - 1] = -p->filter_out;
			p->output = p->filter_out;
			/* El otro extremo */
			p->forward[0] = -backward_out;
			/* Aplicamos la funcin de dispersin del arco al rozar la cuerda */
			add = p->bow_velocity - forward_bow_out - backward_bow_out;
			mul = add * p->bow_table[get_bow_table_index(add)];
			p->forward[p->split] = forward_bow_out + mul;
			p->backward[p->split - 1] = backward_bow_out + mul;
			}
		output[k] = p->output;
		}
	return;
	}

/******************************************************************************
	Abrimos /dev/dsp: 16 bits, little endian, estreo a 44100 Hz.
*/
int audio_open(void) {
	int rate = SAMPLING_RATE;
	int format = AFMT_S16_LE;
	int stereo = 1;
	int fd;

	if ((fd = open("/dev/dsp", O_WRONLY, 0)) == -1) {
		perror("/dev/dsp");
		exit(-1);
		}
	ioctl(fd, SNDCTL_DSP_SPEED, &rate);
	ioctl(fd, SNDCTL_DSP_SETFMT, &format);
	ioctl(fd, SNDCTL_DSP_STEREO, &stereo);
	return fd;
	}

/******************************************************************************
	Reproduce los buffers "left_source" y "right_source" por el dispositivo "fd"
*/
void buffer_play(int fd, float *left_source, float *right_source, float gain) {
	int i, trash;
	short int temp_buffer[BUFFER_SIZE * 2];

	for (i = 0; i < BUFFER_SIZE; i++) {
		/* Canal izquierdo */
		left_source[i] *= gain;
		if (left_source[i] > 1.0)
			left_source[i] = 1.0;
		else if (left_source[i] < -1.0)
			left_source[i] = -1.0;
		temp_buffer[i * 2] = (short) rint(left_source[i] * 32767);
		/* Canal derecho */
		right_source[i] *= gain;
		if (right_source[i] > 1.0)
			right_source[i] = 1.0;
		else if (right_source[i] < -1.0)
			right_source[i] = -1.0;
		temp_buffer[(i * 2) + 1] = (short) rint(right_source[i] * 32767);
		}
	write(fd, temp_buffer, BUFFER_SIZE * 2 * sizeof(short int));
	return;
	}

/*****************************************************************************/
int main(void) {
	int i;
	int dsp;
	bowed_string bs;
	float s[BUFFER_SIZE];

	dsp = audio_open();
	bowed_string_init(&bs, 220, 0.2, 0.6);
	for (i = 0; i < 2000; i++) {
		bowed_string_work(&bs, s);
		buffer_play(dsp, s, s, 1.0);
		}
	close(dsp);
	return 0;
	}
