/*
	Un generador de viento madera.

	Avelino Herrera Morales para "Slo Programadores".

	This software is free software; you can redistribute it and/or modify
	it under the terms of the GNU  General Public License as
	published by the Free Software Foundation; either version 2 of
	the License, or (at your option) any later version.
																																																																																																							 
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU  General Public License for more details.
																																																																																																							 
	You should have received a copy of the GNU  General Public
	License along with this software; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
	02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>


#define  BUFFER_SIZE           256
#define  SAMPLING_RATE         44100
#define  LOWEST_MUSIC_FREQ     20.0
#define  MAX_WAVEGUIDE_SIZE    ((int)((1.0/LOWEST_MUSIC_FREQ) * SAMPLING_RATE/2.0))
#define  REED_TABLE_SIZE       4096
#define  FILTER_CONSTANT       -0.642



typedef struct {
	float freq;
	int waveguide_size;
	float forward[MAX_WAVEGUIDE_SIZE];
	float backward[MAX_WAVEGUIDE_SIZE];
	float reed_table[REED_TABLE_SIZE];
	float mouth_presure;
	float embouchure_offset;
	int start;
	float filter_out;
	float output;
	} woodwind;



int get_reed_table_index(float x) {
	x += 1.0;
	x = (x > 2.0) ? 2.0 : ((x < 0.0) ? 0.0 : x);
	return (int) rint(x * (REED_TABLE_SIZE - 1) / 2.0);
	}


/******************************************************************************
	Inicializamos los campos de la estructura "woodwind".
*/
void woodwind_init(woodwind *p, float freq, float mouth_presure,
                   float embouchure_offset) {
	int i;

	p->waveguide_size = (int) rint((1.0 / freq) * SAMPLING_RATE / 2.0);
	p->mouth_presure = mouth_presure;
	p->embouchure_offset = embouchure_offset;
	for (i = 0; i < REED_TABLE_SIZE; i++) {
		p->reed_table[i] = (float)i * 2.0 / (REED_TABLE_SIZE - 1);
		if (p->reed_table[i] < 1.2)
			p->reed_table[i] = p->reed_table[i] / 1.2;
		 else
			p->reed_table[i] = 1.0;
		}
	p->filter_out = 0;
	p->start = 1;
	return;
	}


/******************************************************************************
	Calculamos BUFFER_SIZE muestras del instrumento.
*/
void woodwind_work(woodwind *p, float *output) {
	int i, k;
	float forward_out, backward_out;
	float add1, add2, mul;

	for (k = 0; k < BUFFER_SIZE; k++) {
		if (p->start) {
			/* Condicin inicial */
			for (i = 0; i < p->waveguide_size; i++) {
				p->forward[i] = 1.0;
				p->backward[i] = -p->forward[i];
				}
			p->filter_out = 0;
			p->start = 0;
			}
		else {
			/* Desplazamos las guas */
			forward_out = p->forward[p->waveguide_size - 1];
			memmove(&(p->forward[1]), &(p->forward[0]),
			        (p->waveguide_size - 1) * sizeof(float));
			backward_out = p->backward[0];
			memmove(&(p->backward[0]), &(p->backward[1]),
			        (p->waveguide_size - 1) * sizeof(float));
			/* La campana: filtro paso-bajo de reflexin y filtro paso-alto de
			   salida */
			p->filter_out = (forward_out * (1 + FILTER_CONSTANT)) -
			                (FILTER_CONSTANT * p->filter_out);
			p->backward[p->waveguide_size - 1] = -p->filter_out;
			p->output = forward_out - p->filter_out;
			/* La boca */
			add1 = p->mouth_presure - p->embouchure_offset - backward_out;
			mul = add1 * p->reed_table[get_reed_table_index(add1)];
			add2 = p->mouth_presure - mul;
			p->forward[0] = add2;
			}
		output[k] = p->output;
		}
	return;
	}


int audio_open(void) {
	int rate = SAMPLING_RATE;
	int format = AFMT_S16_LE;
	int stereo = 1;
	int fd;

	if ((fd = open("/dev/dsp", O_WRONLY, 0)) == -1) {
		perror("/dev/dsp");
		exit(-1);
		}
	ioctl(fd, SNDCTL_DSP_SPEED, &rate);
	ioctl(fd, SNDCTL_DSP_SETFMT, &format);
	ioctl(fd, SNDCTL_DSP_STEREO, &stereo);
	return fd;
	}


void buffer_play(int fd, float *left_source, float *right_source, float gain) {
	int i, trash;
	short int temp_buffer[BUFFER_SIZE * 2];

	for (i = 0; i < BUFFER_SIZE; i++) {
		left_source[i] *= gain;
		if (left_source[i] > 1.0)
			left_source[i] = 1.0;
		else if (left_source[i] < -1.0)
			left_source[i] = -1.0;
		temp_buffer[i * 2] = (short) rint(left_source[i] * 32767);
		right_source[i] *= gain;
		if (right_source[i] > 1.0)
			right_source[i] = 1.0;
		else if (right_source[i] < -1.0)
			right_source[i] = -1.0;
		temp_buffer[(i * 2) + 1] = (short) rint(right_source[i] * 32767);
		}
	write(fd, temp_buffer, BUFFER_SIZE * 2 * sizeof(short int));
	return;
	}

/*****************************************************************************/
int main(void) {
	int i;
	int dsp;
	woodwind ww;
	float s[BUFFER_SIZE];

	dsp = audio_open();
	woodwind_init(&ww, 880, 0.28, 0.0);
	for (i = 0; i < 2000; i++) {
		woodwind_work(&ww, s);
		buffer_play(dsp, s, s, 4.0);
		}
	close(dsp);
	return 0;
	}
